#include <unistd.h>

#include <CoreFoundation/CoreFoundation.h>

#include <IOKit/IOKitLib.h>
#include <IOKit/IOCFPlugIn.h>
#include <IOKit/usb/IOUSBLib.h>
#include <mach/mach.h>
#include "hex2c.h"

#define ACCES_VENDOR_ID		0x1605

#define USB_DIO_32_ID_RAW	0x0001
#define USB_DIO_32_ID_DEV	0x8001

#define USB_DA12_8A_REVA_RAW	0x4001
#define USB_DA12_8A_REVA_DEV	0xC001

#define USB_DA12_8A_RAW	0x4002
#define USB_DA12_8A_DEV	0xC002

#define USB_IIRO_16_RAW	0x0010
#define USB_IIRO_16_DEV 0x8010

#define USB_CTR_15_RAW	0x0020
#define USB_CTR_15_DEV	0x8020

#define USB_IIRO4_2SM_RAW	0x0030
#define USB_IIRO4_2SM_DEV	0x8030

#define	k8051_USBCS		0xe600

extern INTEL_HEX_RECORD ctr15Hex[];
extern INTEL_HEX_RECORD da128ArAHex[];
extern INTEL_HEX_RECORD dio32Hex[];
extern INTEL_HEX_RECORD iiro16Hex[];
extern INTEL_HEX_RECORD iiro42smHex[];

static IONotificationPortRef	gNotifyPort;


static io_iterator_t		gIiro16AddedIter;
static io_iterator_t		gDio32AddedIter;
static io_iterator_t		gDA128ArAAddedIter;
static io_iterator_t		gCtr15AddedIter;
static io_iterator_t		gIiro42smAddedIter;



static io_iterator_t		gIiro16RemovedIter;
static io_iterator_t		gDio32RemovedIter;
static io_iterator_t		gDA128ArARemovedIter;
static io_iterator_t		gCtr15RemovedIter;
static io_iterator_t		gIiro42smRemovedIter;

bool devProgged;



IOReturn ConfigureACCESDevice(IOUSBDeviceInterface **dev)
{
    UInt8				numConf;
    IOReturn				kr;
    IOUSBConfigurationDescriptorPtr	confDesc;
    
    kr = (*dev)->GetNumberOfConfigurations(dev, &numConf);
    if (!numConf)
        return -1;
    
    // get the configuration descriptor for index 0
    kr = (*dev)->GetConfigurationDescriptorPtr(dev, 0, &confDesc);
    if (kr)
    {
        printf("\tunable to get config descriptor for index %d (err = %08x)\n", 0, kr);
        return -1;
    }
    kr = (*dev)->SetConfiguration(dev, confDesc->bConfigurationValue);
    if (kr)
    {
        printf("\tunable to set configuration to value %d (err=%08x)\n", 0, kr);
        return -1;
    }
    
    return kIOReturnSuccess;
}

IOReturn USBByteWrite (IOUSBDeviceInterface **dev, UInt16 address, UInt16 count, UInt8 writeBuffer[])
{
    IOUSBDevRequest 		request;
    
    request.bmRequestType = USBmakebmRequestType(kUSBOut, kUSBVendor, kUSBDevice);
    request.bRequest = 0xa0;
    request.wValue = address;
    request.wIndex = 0;
    request.wLength = count;
    request.pData = writeBuffer;

    return (*dev)->DeviceRequest(dev, &request);
}

IOReturn programDev (IOUSBDeviceInterface **dev, UInt16 RawID)
{
	UInt8		writeVal;
	IOReturn	retVal;
	INTEL_HEX_RECORD current;
	PINTEL_HEX_RECORD accesHex;
	int count = 0;
	
	
	
	switch(RawID)
	{
	case USB_DIO_32_ID_RAW:		accesHex = dio32Hex; break;
	case USB_DA12_8A_REVA_RAW:	accesHex = da128ArAHex; break;
	case USB_IIRO_16_RAW:		accesHex = iiro16Hex; break;
	case USB_CTR_15_RAW:		accesHex = ctr15Hex; break;
	case USB_IIRO4_2SM_RAW:		accesHex = iiro42smHex; break;
	default: printf("Unknown Product ID: %04x\n", RawID);
		return kIOReturnUnsupported;
	}
	
	

	writeVal = 1;
	
	retVal = USBByteWrite(dev, k8051_USBCS, 1, &writeVal);
	
	if (retVal != kIOReturnSuccess)
	{
		printf ("Could not write the reset to the board");
		return retVal;
	}
	
	do
	{
		current = accesHex[count];
		count++;
		
		if (current.Type != 1)
		{
			retVal = USBByteWrite(dev, current.Address, current.Length, current.Data);
			
			if (retVal != kIOReturnSuccess)
			{
				printf("Error while performing write to USB address %x\n", current.Address);
				return retVal;
			}
		
		}
	
	}while (current.Type != 1);
	
	//now that we've put the firmware in the board we need to tell it to re-enumerate
	writeVal = 0;
	
	retVal = USBByteWrite(dev, k8051_USBCS, 1, &writeVal);
	
	return retVal;

}



void rawAdded(void *refCon, io_iterator_t iterator)
{
		kern_return_t			kernRetVal;
	io_service_t			usbDevice;
	IOCFPlugInInterface		**plugInInterface;
	IOUSBDeviceInterface	**dev = NULL;
	HRESULT					hRetVal;
	SInt32					score;
	UInt16					vendor;
	UInt16					product;
//	UInt16					release;
	int						exclusiveCheck = 0;
	
	while ( (usbDevice = IOIteratorNext(iterator)) )
	{
		kernRetVal = IOCreatePlugInInterfaceForService(usbDevice, kIOUSBDeviceUserClientTypeID, 
														kIOCFPlugInInterfaceID, &plugInInterface, &score);
														
		//we now have the plug in Interface so we don't need the device object
		
		kernRetVal = IOObjectRelease(usbDevice);
		
		if (( kernRetVal != kIOReturnSuccess) || (plugInInterface == 0))
		{
			printf("Could not make plugin Interface");
			continue;
		}
		
		// now we need to create the device interface, and ditch the plug in interface
		
		hRetVal = (*plugInInterface)->QueryInterface(plugInInterface, CFUUIDGetUUIDBytes(kIOUSBDeviceInterfaceID),
					(LPVOID)&dev);
					
		IODestroyPlugInInterface(plugInInterface);
		
		if ((hRetVal != 0) || (dev == 0))
		{
			printf("Could not create the device Interface");
			continue;
		}
		
		//check the vendor ID and the Product ID. May have to add Release Number Checking later
		kernRetVal = (*dev) -> GetDeviceVendor(dev, &vendor);
		kernRetVal = (*dev) -> GetDeviceProduct ( dev, &product);
		
		if (vendor != ACCES_VENDOR_ID)
		{
			printf("Wrong Vendor ID received\n");
			continue;
		}
		
		//now that we have the device we are looking for we need to see if we can get exclusive access
		// to it
		
		do
		{
			kernRetVal = (*dev) -> USBDeviceOpen(dev);
			
			if (kernRetVal == kIOReturnExclusiveAccess) //we can't get it right this moment so we will try a while
			{
				exclusiveCheck++;
				sleep(1);
			}
		}while ( (kernRetVal == kIOReturnExclusiveAccess) && (exclusiveCheck < 5));
		
		if (kernRetVal != kIOReturnSuccess)
		{
			printf("Could not open device");
			(void)(*dev)->Release(dev);
			continue;
		}
		
		kernRetVal = ConfigureACCESDevice(dev);
		
		if (kernRetVal != kIOReturnSuccess)
		{
			(void) (*dev)-> USBDeviceClose(dev);
			(void) (*dev) -> Release(dev);
			continue;
		}
		
		kernRetVal = programDev(dev, product);
		
		if (kernRetVal != kIOReturnSuccess)
		{
			printf("Programming device failed.\n");
		}
		else
		{
			switch(product)
			{
			case USB_DIO_32_ID_RAW:
				printf("USB-DIO-32 programmed. Product ID: %4x\n", USB_DIO_32_ID_DEV);
				break;
			case USB_DA12_8A_REVA_RAW:
				printf("USB-DA12-8A programmed. Product ID: %4x\n", USB_DA12_8A_REVA_DEV);
				break;
			case USB_IIRO_16_RAW:
				printf("USB-IIRO-16 programmed. Product ID: %4x\n", USB_IIRO_16_DEV);
				break;
			case USB_CTR_15_RAW:
				printf("USB-CTR-15 programmed. Product ID: %4x\n", USB_CTR_15_DEV);
				break;
			case USB_IIRO4_2SM_RAW:
				printf("USB-IIRO4-2SM programmed. Product ID: %4x\n", USB_IIRO4_2SM_DEV);
				break;
			default:
				printf("Unknown device programmed. Abandon all hope ye who reach this line of code.\n");
				break;
			};
			
			
			
			devProgged = true;
		}
		
		kernRetVal = (*dev)-> USBDeviceClose(dev);
		kernRetVal = (*dev)-> Release(dev);
		
		
	}
}

void iiro16Added(void *refCon, io_iterator_t iterator)
{
	rawAdded(refCon, iterator);
}

void dio32Added(void *refCon, io_iterator_t iterator)
{
	rawAdded(refCon, iterator);
}

void da128ArAAdded(void *refCon, io_iterator_t iterator)
{
	rawAdded(refCon, iterator);
}

void ctr15Added(void *refCon, io_iterator_t iterator)
{
	rawAdded(refCon, iterator);
}

void iiro42smAdded(void *refCon, io_iterator_t iterator)
{
	rawAdded(refCon, iterator);
}

void rawRemoved(void *refCon, io_iterator_t iterator)
{
	kern_return_t	kr;
    io_service_t	obj;
    
    while ( (obj = IOIteratorNext(iterator)) )
    {
        //printf("Raw device removed.\n");
        kr = IOObjectRelease(obj);
    }
	
	if (devProgged == true)
	{
		exit(0);
	}
}

void iiro16Removed(void *refCon, io_iterator_t iterator)
{
	rawRemoved(refCon, iterator);
}

void dio32Removed(void *refCon, io_iterator_t iterator)
{
	rawRemoved(refCon, iterator);
}

void da128ArARemoved(void *refCon, io_iterator_t iterator)
{
	rawRemoved(refCon, iterator);
}

void ctr15Removed(void *refCon, io_iterator_t iterator)
{
	rawRemoved(refCon, iterator);
}

void iiro42smRemoved(void *refCon, io_iterator_t iterator)
{
	rawRemoved(refCon, iterator);
}

int main (int argc, const char * argv[]) {
    // insert code here...
    
	mach_port_t				ioPort;
	CFMutableDictionaryRef	dictionary;
	SInt32					usbVendor, usbDevID;
	CFRunLoopSourceRef		runLoopSource;
	
	kern_return_t			retVal;
	
	retVal = IOMasterPort(MACH_PORT_NULL, &ioPort);
	
	devProgged = false;
	
	if ((retVal != 0) || (ioPort == 0))
	{
		printf("Could not create port.\n");
		exit(1);
	}
	
	dictionary = IOServiceMatching(kIOUSBDeviceClassName);
	
	if (dictionary == 0)
	{
		printf("Could not create Dictionary");
		exit(2);
	}
	
	usbVendor = ACCES_VENDOR_ID;
	
	CFDictionarySetValue(dictionary, CFSTR(kUSBVendorID), 
						 CFNumberCreate(kCFAllocatorDefault, kCFNumberSInt32Type, &usbVendor));
						 
	usbDevID = USB_DIO_32_ID_RAW;					 
						 
	CFDictionarySetValue(dictionary, CFSTR(kUSBProductID),
							CFNumberCreate(kCFAllocatorDefault, kCFNumberSInt32Type, &usbDevID));
							
	//now that we have the dictionary we can request to be notified when events occur on 
	//devices that match the dictionary.
	
	gNotifyPort = IONotificationPortCreate(ioPort);
	runLoopSource = IONotificationPortGetRunLoopSource(gNotifyPort);
	
	CFRunLoopAddSource(CFRunLoopGetCurrent(), runLoopSource, kCFRunLoopDefaultMode);
	
	//we get one reference to the dictionary by default, but each notification handler consumes a reference
	//so we need more.
	
	dictionary = (CFMutableDictionaryRef) CFRetain(dictionary);
	dictionary = (CFMutableDictionaryRef) CFRetain(dictionary);
	
	IOServiceAddMatchingNotification(gNotifyPort,
										kIOFirstMatchNotification,
										dictionary,
										dio32Added,
										NULL,
										&gDio32AddedIter);
										


	IOServiceAddMatchingNotification(gNotifyPort,
										kIOTerminatedNotification,
										dictionary,
										dio32Removed,
										NULL,
										&gDio32RemovedIter);
										

	dio32Added(NULL, gDio32AddedIter);
	dio32Removed(NULL, gDio32RemovedIter);

	
	//now for the USB-DA12-8RevA
	
	usbDevID = USB_DA12_8A_REVA_RAW;
	
	CFDictionarySetValue(dictionary, CFSTR(kUSBProductID),
							CFNumberCreate(kCFAllocatorDefault, kCFNumberSInt32Type, &usbDevID));
							
	//now we need to get two more references for the new notifications we are going to request
	dictionary = (CFMutableDictionaryRef) CFRetain(dictionary);
	dictionary = (CFMutableDictionaryRef) CFRetain(dictionary);
	
	IOServiceAddMatchingNotification(gNotifyPort,
										kIOFirstMatchNotification,
										dictionary,
										da128ArAAdded,
										NULL,
										&gDA128ArAAddedIter);
										
	IOServiceAddMatchingNotification(gNotifyPort,
										kIOTerminatedNotification,
										dictionary,
										da128ArARemoved,
										NULL,
										&gDA128ArARemovedIter);
										
	da128ArAAdded(NULL, gDA128ArAAddedIter);
	da128ArARemoved(NULL, gDA128ArARemovedIter);
										
										
	//USB-IIRO-16
	usbDevID = USB_IIRO_16_RAW;
	
	CFDictionarySetValue(dictionary, CFSTR(kUSBProductID),
							CFNumberCreate(kCFAllocatorDefault, kCFNumberSInt32Type, &usbDevID));
							
	dictionary = (CFMutableDictionaryRef) CFRetain(dictionary);
	dictionary = (CFMutableDictionaryRef) CFRetain(dictionary);
	
	IOServiceAddMatchingNotification(gNotifyPort,
										kIOFirstMatchNotification,
										dictionary,
										iiro16Added,
										NULL,
										&gIiro16AddedIter);
										
	
										
	IOServiceAddMatchingNotification(gNotifyPort,
										kIOTerminatedNotification,
										dictionary,
										iiro16Removed,
										NULL,
										&gIiro16RemovedIter);
										
	iiro16Added(NULL, gIiro16AddedIter);
	iiro16Removed(NULL, gIiro16RemovedIter);
	
	//for IIRO4-2SM
	usbDevID = USB_IIRO4_2SM_RAW;
	
	CFDictionarySetValue(dictionary, CFSTR(kUSBProductID),
							CFNumberCreate(kCFAllocatorDefault, kCFNumberSInt32Type, &usbDevID));
							
	dictionary = (CFMutableDictionaryRef) CFRetain(dictionary);
	dictionary = (CFMutableDictionaryRef) CFRetain(dictionary);	
	
	IOServiceAddMatchingNotification(gNotifyPort,
										kIOFirstMatchNotification,
										dictionary,
										iiro42smAdded,
										NULL,
										&gIiro42smAddedIter);
										
	IOServiceAddMatchingNotification(gNotifyPort,
										kIOTerminatedNotification,
										dictionary,
										iiro42smRemoved,
										NULL,
										&gIiro42smRemovedIter);
										
	iiro42smAdded(NULL, gIiro42smAddedIter);
	iiro42smRemoved(NULL, gIiro42smRemovedIter);
	
	
	/*************************************************************
	//Any new devices should be added here, above the USB_CTR_15
	//and follow the pattern of the device above it. The reason for this is that there
	//are slight differences between the last device and the device before it, and it the program
	//will probably give a bus error when run if the USB_CTR_15 example is used or if the new device
	//is placed after it.
	*************************************************************/
	

	//now for USB_CTR_15
	usbDevID = USB_CTR_15_RAW;
	
	CFDictionarySetValue(dictionary, CFSTR(kUSBProductID),
							CFNumberCreate(kCFAllocatorDefault, kCFNumberSInt32Type, &usbDevID));
	
	dictionary = (CFMutableDictionaryRef) CFRetain(dictionary);
	
	
	IOServiceAddMatchingNotification(gNotifyPort,
										kIOFirstMatchNotification,
										dictionary,
										ctr15Added,
										NULL,
										&gCtr15AddedIter);
										
	IOServiceAddMatchingNotification(gNotifyPort,
										kIOTerminatedNotification,
										dictionary,
										ctr15Removed,
										NULL,
										&gCtr15RemovedIter);
										

										
	//call each one of the handlers to catch any already received messages and to "arm" them.
	ctr15Added(NULL, gCtr15AddedIter); //call it once to catch already attached devices and arm the notification
	rawRemoved(NULL, gCtr15RemovedIter); //call it once to arm the notification											
										
	
	//done with port
	mach_port_deallocate(mach_task_self(), ioPort);
	ioPort = 0;
	
	CFRunLoopRun(); //this will keep going until something calls CFRunLoopStop()
	
    return 0;
}