/*
 *  aiousb.c
 *  aiousb
 *
 *  Created by Jay Dolan on 3/14/06.
 *  This source code may be reproduced, in whole or in part. 
 *  It may be modified in any way. The only request made by ACCES is that any product
 *  using this code must interact with ACCES hardware.
 *
 */

#include "aiousb.h"
#define MAX_DEVICES		32
#define NUM_PRODUCTS	10

#define EXPORT __attribute__((visibility("default")))

typedef struct deviceStruct
{
	UInt16				prodID;
	UInt8				DIOBytes;
	UInt8				counters;
	bool				connected;
	io_iterator_t		removedIter;
	USBDeviceAddress	addr;
	IOUSBDeviceInterface	**interface;
} DEVICE;

static DEVICE deviceList[MAX_DEVICES];

static IONotificationPortRef	gNotifyPort;
static io_iterator_t			gAddIter;
static mach_port_t				ioPort;
static CFRunLoopSourceRef		runLoopSource;

bool	gDevFound;
UInt8	gNumDevices;

//function prototypes
void devRemoved(void *refCon, io_iterator_t iterator);

EXPORT
SInt8 CTR_8254Read (SInt8 deviceIndex, UInt8 blockIndex, UInt8 counterIndex, UInt16 *value)
{
	IOUSBDevRequest request;
	IOUSBDeviceInterface **dev;
	IOReturn retval;
	
	devRemoved(NULL, deviceList[deviceIndex].removedIter);
	
	if ((deviceIndex < 0) || (deviceIndex > (gNumDevices - 1)))
		return INVALID_DEVICE_INDEX;
	
	if (deviceList[deviceIndex].connected == false)
		return DEVICE_NOT_CONNECTED;
	
	if (deviceList[deviceIndex].counters == 0)
		return OPERATION_NOT_SUPPORTED;	
	
	if (blockIndex >= deviceList[deviceIndex].counters)
		return INVALID_BLOCK_INDEX;
	
	if (counterIndex > 2)
		return INVALID_COUNTER_INDEX;
	
	request.bmRequestType = USBmakebmRequestType (kUSBIn, kUSBVendor, kUSBDevice);
	request.bRequest = 0x20;
	request.wValue = (counterIndex << 8) | blockIndex;  
	request.wIndex = 0;
	request.wLength = 2;
	request.pData = value;
	
	dev = deviceList[deviceIndex].interface;
	
	retval = (*dev) -> DeviceRequest(dev, &request);
	
	*value = USBToHostWord(*value);
	
	return retval;
}

EXPORT
SInt8 CTR_8254ModeLoad (SInt8 deviceIndex, UInt8 blockIndex, UInt8 counterIndex, UInt8 mode, UInt16 value)
{
	UInt8 modeByte;
	IOUSBDeviceInterface **dev;
	IOUSBDevRequest request;
	
	devRemoved(NULL, deviceList[deviceIndex].removedIter);
	
	if ((deviceIndex < 0) || (deviceIndex > (gNumDevices - 1)))
		return INVALID_DEVICE_INDEX;
	
	if (deviceList[deviceIndex].connected == false)
		return DEVICE_NOT_CONNECTED;
	
	if (deviceList[deviceIndex].counters == 0)
		return OPERATION_NOT_SUPPORTED;	
	
	if (blockIndex >= deviceList[deviceIndex].counters)
		return INVALID_BLOCK_INDEX;
	
	if (counterIndex > 2)
		return INVALID_COUNTER_INDEX;
	
	modeByte = (counterIndex << 6) | 0x30 | (mode << 1);
	
	request.bmRequestType = USBmakebmRequestType (kUSBOut, kUSBVendor, kUSBDevice);
	request.bRequest = 0x23;
	request.wValue =   (modeByte << 8) | blockIndex; //HostToUSBWord ((modeByte << 8) | blockIndex);
	request.wIndex = value; //HostToUSBWord (value);
	request.wLength = 0;
	request.pData = NULL;
	
	dev = deviceList[deviceIndex].interface;
	
	return (*dev) -> DeviceRequest(dev, &request);
	
	
}

EXPORT
SInt8 CTR_8254Mode (SInt8 deviceIndex, UInt8 blockIndex, UInt8 counterIndex, UInt8 mode)
{
	UInt8 modeByte;
	IOUSBDevRequest request;
	IOUSBDeviceInterface **dev;
	
	devRemoved(NULL, deviceList[deviceIndex].removedIter);
	
	if ((deviceIndex < 0) || (deviceIndex > (gNumDevices - 1)))
		return INVALID_DEVICE_INDEX;
	
	if (deviceList[deviceIndex].connected == false)
		return DEVICE_NOT_CONNECTED;
	
	if (deviceList[deviceIndex].counters == 0)
		return OPERATION_NOT_SUPPORTED;	
	
	if (blockIndex >= deviceList[deviceIndex].counters)
		return INVALID_BLOCK_INDEX;
	
	if (counterIndex > 2)
		return INVALID_COUNTER_INDEX;
	
	modeByte = (counterIndex << 6) | 0x30 | (mode << 1);
	
	request.bmRequestType = USBmakebmRequestType (kUSBOut, kUSBVendor, kUSBDevice);
	request.bRequest = 0x21;
	request.wValue = (modeByte << 8) | blockIndex;
	request.wIndex = 0;
	request.wLength = 0;
	request.pData = NULL;
	
	dev = deviceList[deviceIndex].interface;
	
	return (*dev) -> DeviceRequest(dev, &request);
}

EXPORT
SInt8 DIO_Read1 (SInt8 deviceIndex, UInt8 bitIndex, bool *data)
{
	UInt8 *pData;
	SInt8 byteIndex;
	SInt8 shift;
	
	if ((deviceIndex < 0) || (deviceIndex > (gNumDevices - 1)))
		return INVALID_DEVICE_INDEX;
	
	if (deviceList[deviceIndex].connected == false)
		return DEVICE_NOT_CONNECTED;
	
	if (deviceList[deviceIndex].DIOBytes == 0)
		return OPERATION_NOT_SUPPORTED;
	
	if (bitIndex > (deviceList[deviceIndex].DIOBytes * 8 - 1))
		return INVALID_BIT_INDEX;
	
	pData = malloc(deviceList[deviceIndex].DIOBytes);
	
	DIO_ReadAll(deviceIndex, pData);
	
	byteIndex = bitIndex / 8; 
	
	shift = (bitIndex % 8);
	
	*data = pData[byteIndex] & (1 << shift);
	
	
	free(pData);
	
	return 0;
	
}

EXPORT
SInt8 DIO_Read8 (SInt8 deviceIndex, UInt8 byteIndex, UInt8 *data)
{
	UInt8 *pData;
	
	if ((deviceIndex < 0) || (deviceIndex > (gNumDevices - 1)))
		return INVALID_DEVICE_INDEX;
	
	if (deviceList[deviceIndex].connected == false)
		return DEVICE_NOT_CONNECTED;
	
	if (deviceList[deviceIndex].DIOBytes == 0)
		return OPERATION_NOT_SUPPORTED;
	
	if (byteIndex > (deviceList[deviceIndex].DIOBytes - 1))
		return INVALID_BYTE_INDEX;
	
	pData = malloc(deviceList[deviceIndex].DIOBytes);
	
	DIO_ReadAll(deviceIndex, pData);
	
	*data = pData[byteIndex];
	
	
	free(pData);
	
	return 0;
	
}

EXPORT
SInt8 DIO_Write1 (SInt8 deviceIndex, UInt8 bitIndex, bool *data)
{
	UInt8 *pData;
	SInt8 byteIndex;
	SInt8 retVal;
	SInt8 shift;
	
	if ((deviceIndex < 0) || (deviceIndex > (gNumDevices - 1)))
		return INVALID_DEVICE_INDEX;
	
	if (deviceList[deviceIndex].connected == false)
		return DEVICE_NOT_CONNECTED;
	
	if (deviceList[deviceIndex].DIOBytes == 0)
		return OPERATION_NOT_SUPPORTED;
	
	if (bitIndex > (deviceList[deviceIndex].DIOBytes * 8 - 1))
		return INVALID_BIT_INDEX;
	
	pData = malloc(deviceList[deviceIndex].DIOBytes);
	
	DIO_ReadAll(deviceIndex, pData);
	
	byteIndex = bitIndex / 8 ; 
	
	shift = bitIndex % 8;
	
	
	//sets the bit to 1 or 0 depending on weather data is true or false
	pData[byteIndex] = *data ? pData[byteIndex] | (1 << shift) :  pData[byteIndex] & (0xFF ^ (1 << shift));
	
	retVal = DIO_WriteAll(deviceIndex, pData);
	
	free(pData);
	
	return retVal;
	
}

EXPORT
SInt8 DIO_Write8 (SInt8 deviceIndex, UInt8 byteIndex, UInt8 *data)
{
	UInt8 *pData;
	SInt8 retVal;
	
	if ((deviceIndex < 0) || (deviceIndex > (gNumDevices - 1)))
		return INVALID_DEVICE_INDEX;
	
	if (deviceList[deviceIndex].connected == false)
		return DEVICE_NOT_CONNECTED;
	
	if (deviceList[deviceIndex].DIOBytes == 0)
		return OPERATION_NOT_SUPPORTED;
	
	if (byteIndex > (deviceList[deviceIndex].DIOBytes - 1))
		return INVALID_BYTE_INDEX;

	pData = malloc(deviceList[deviceIndex].DIOBytes);
	
	DIO_ReadAll(deviceIndex, pData);
	
	pData[byteIndex] = *data;
	
	retVal =  DIO_WriteAll(deviceIndex, pData);
	
	free(pData);
	
	return retVal;
	
}

EXPORT
SInt8 DIO_ReadAll (SInt8 deviceIndex, UInt8 *pData)
{
	IOUSBDevRequest request;
	IOUSBDeviceInterface **dev;
	IOReturn retval;
	
	devRemoved(NULL, deviceList[deviceIndex].removedIter);
	
	if ((deviceIndex < 0) || (deviceIndex > (gNumDevices - 1)))
		return INVALID_DEVICE_INDEX;
	
	if (deviceList[deviceIndex].connected == false)
		return DEVICE_NOT_CONNECTED;
	
	if (deviceList[deviceIndex].DIOBytes == 0)
		return OPERATION_NOT_SUPPORTED;
	
	request.bmRequestType = USBmakebmRequestType (kUSBIn, kUSBVendor, kUSBDevice);
	request.bRequest = 0x11;
	request.wValue = 0;
	request.wIndex = 0;
	request.wLength = deviceList[deviceIndex].DIOBytes;
	request.pData = pData;
	
	dev = deviceList[deviceIndex].interface;
	
	retval = (*dev) -> DeviceRequest(dev, &request);
	
	if (retval == kIOReturnSuccess)
		return 0;
	else
		return -128;
}

EXPORT
SInt8 DIO_WriteAll (SInt8 deviceIndex, UInt8 *pData)
{
	IOUSBDevRequest	request;
	IOUSBDeviceInterface **dev;
	IOReturn	retval;
	
	devRemoved(NULL, deviceList[deviceIndex].removedIter);
	
	if ((deviceIndex < 0) || (deviceIndex > (gNumDevices - 1)))
	{
		return INVALID_DEVICE_INDEX;
	}
	
	if (deviceList[deviceIndex].connected == false)
	{
		return DEVICE_NOT_CONNECTED;
	}
	
	//need to check and see if it is a device that supports DIO...
	if (deviceList[deviceIndex].DIOBytes == 0)
		return OPERATION_NOT_SUPPORTED;
	
	
	
	request.bmRequestType = USBmakebmRequestType (kUSBOut, kUSBVendor, kUSBDevice);
	request.bRequest = 0x10;
	request.wValue = 0;
	request.wIndex = 0;
	request.wLength = deviceList[deviceIndex].DIOBytes;
	request.pData = pData;
	
	dev = deviceList[deviceIndex].interface;
	
	retval = (*dev) -> DeviceRequest(dev, &request);
	
	if (retval == kIOReturnSuccess)
		return 0;
	else
		return -128;
	
	
}

EXPORT
SInt8 queryDeviceInfo(SInt8 deviceIndex, UInt16 *productID, UInt16 nameSize, char *name, UInt8 *DIOBytes, UInt8 *counters)
{
	UInt8 bytes, ctrs;

	if ((deviceIndex < 0) || (deviceIndex > (gNumDevices - 1)))
	{
		return INVALID_DEVICE_INDEX;
	}
	
	if (deviceList[deviceIndex].connected == false)
	{
		return DEVICE_NOT_CONNECTED;
	}
	
	if (productID != NULL)
		*productID = deviceList[deviceIndex].prodID;
	
	
	switch(deviceList[deviceIndex].prodID)
	{
	case USB_DIO_32_ID_DEV:
		bytes = 4;
		ctrs = 3;
		strncpy(name , "USB-DIO-32" ,nameSize);
		break;
	case USB_DA12_8A_REVA_DEV:
		bytes = 0;
		ctrs = 0;
		strncpy(name, "USB-DA12-8A", nameSize);
		break;
	case USB_IIRO_16_DEV:		
		bytes = 4;
		ctrs = 0;
		strncpy(name, "USB-IIRO-16", nameSize);
		break;
	case USB_II_16_DEV:
		bytes = 4;
		ctrs = 0;
		strncpy(name, "USB-II-16", nameSize);
		break;
	case USB_RO_16_DEV:
		bytes = 4;
		ctrs = 0;
		strncpy(name, "USB-RO-16", nameSize);
		break;
	case USB_IIRO_8_DEV:
		bytes = 4;
		ctrs = 0;
		strncpy(name, "USB-IIRO-8", nameSize);
		break;
	case USB_II_8_DEV:
		bytes = 4;
		ctrs = 0;
		strncpy(name, "USB-II-8", nameSize);
		break;
	case USB_RO_8_DEV:
		bytes = 4;
		ctrs = 0;
		strncpy(name, "USB-RO-8", nameSize);			
	case USB_CTR_15_DEV:
		ctrs = 5;
		bytes = 0;
		strncpy(name, "USB-CTR-15", nameSize);
		break;
	case USB_IIRO4_2SM_DEV:
		bytes = 2;
		ctrs = 0;
		strncpy(name, "USB-IIRO4-2SM", nameSize);
		break;
	}
	
	if (DIOBytes != NULL)
		*DIOBytes = bytes;
		
	if (counters != NULL)
		*counters = ctrs;
		
	return 0;
}

EXPORT
SInt8 dioConfig(SInt8 deviceIndex, bool tristate, UInt8 outMask, UInt8 *pBytes)
{
	IOUSBDevRequest		request;
	UInt8 byteArray[6];
	IOUSBDeviceInterface **dev;
	IOReturn retval;
	int count;
	
	if ((deviceIndex < 0) || (deviceIndex > (gNumDevices - 1)))
	{
		return INVALID_DEVICE_INDEX;
	}
	
	if (deviceList[deviceIndex].connected == false)
	{
		return DEVICE_NOT_CONNECTED;
	}
	
	if (deviceList[deviceIndex].DIOBytes == 0)
		return OPERATION_NOT_SUPPORTED;
	
	request.bmRequestType = USBmakebmRequestType (kUSBOut, kUSBVendor, kUSBDevice);
	request.bRequest = 0x12;
	request.wValue = (tristate ? 1 : 0);
	request.wIndex = 0;
	request.wLength = deviceList[deviceIndex].DIOBytes + 1;
	
	
	for (count = 0; count < deviceList[deviceIndex].DIOBytes; count++)
	{
		byteArray[count] = pBytes[count];
	}
	byteArray[count] = outMask;
	byteArray[count + 1] = 0;
	
	
	request.pData = byteArray;
	
	dev = deviceList[deviceIndex].interface;
	
	retval = (*dev) -> DeviceRequest(dev, &request);
	
	if (retval == kIOReturnSuccess)
		return 0;
	else
		return -128;
}

EXPORT
void devAdded(void *refCon, io_iterator_t iterator)
{
	io_service_t			obj;
	IOCFPlugInInterface		**plugInInterface;
	IOUSBDeviceInterface	**dev = NULL;
	SInt32					score;
	USBDeviceAddress		currentAddr;
	UInt8					count;
	bool					dupe = false;

	if (obj = IOIteratorNext(iterator))
	{
		IOCreatePlugInInterfaceForService(obj, kIOUSBDeviceUserClientTypeID, kIOCFPlugInInterfaceID,
											&plugInInterface, &score);
											
		IOObjectRelease(obj);
		
		(*plugInInterface) ->QueryInterface(plugInInterface, CFUUIDGetUUIDBytes(kIOUSBDeviceInterfaceID),
								(LPVOID)&dev);
								
		IODestroyPlugInInterface(plugInInterface);
		
		(*dev) ->GetDeviceAddress(dev, &currentAddr);
		
		for (count = 0; count < gNumDevices; count++)
		{
			if (deviceList[count].addr == currentAddr)
				dupe = true;
		}
		
		if (!dupe)
		{
			deviceList[gNumDevices].addr = currentAddr;
			deviceList[gNumDevices].interface = dev;
			gDevFound = true;
		}
		
	}
}


void devRemoved(void *refCon, io_iterator_t iterator)
{
	bool done = false;
	int count = 0;
	io_service_t obj;
	
	if ( obj = IOIteratorNext(iterator))
	{
		do
		{
			if (deviceList[count].removedIter == iterator)
			{
				deviceList[count].prodID = 0;
				deviceList[count].connected = false;
				IOObjectRelease(obj);
				done = true;
			}
			count ++;
		}while ((done == false) && (count < gNumDevices));
	}
}

EXPORT
SInt8 queryDevices(void)
{
	//mach_port_t				ioPort;
	CFMutableDictionaryRef	dictionary;
	SInt32					usbVendor, usbProdID;
	//CFRunLoopSourceRef		runLoopSource;
	
	UInt8					count;
	
	kern_return_t			retVal;
	gNumDevices = 0;
	
	retVal = IOMasterPort(MACH_PORT_NULL, &ioPort);
	
	if ((retVal != 0) || (ioPort == 0))
	{
		return COULD_NOT_CREATE_MASTER_PORT;
	}
	
	dictionary = IOServiceMatching(kIOUSBDeviceClassName);
	
	if (dictionary == 0)
	{
		return COULD_NOT_CREATE_DICTIONARY;
	}
	
	usbVendor = ACCES_VENDOR_ID;
	
	CFDictionarySetValue(dictionary, CFSTR(kUSBVendorID),
							CFNumberCreate(kCFAllocatorDefault, kCFNumberSInt32Type, &usbVendor));
							
	gNotifyPort = IONotificationPortCreate(ioPort);
	runLoopSource = IONotificationPortGetRunLoopSource(gNotifyPort);

	for (count = 0; count < MAX_DEVICES; count++)
		deviceList[count].addr = 0;
	
	for (count = 0; count < NUM_PRODUCTS; count++)
	{
		switch(count)
		{
		case 0: usbProdID = USB_DIO_32_ID_DEV; break;
		case 1: usbProdID = USB_DA12_8A_REVA_DEV; break;
		case 2: usbProdID = USB_IIRO_16_DEV; break;
		case 3: usbProdID = USB_CTR_15_DEV; break;
		case 4: usbProdID = USB_IIRO4_2SM_DEV; break;
		case 5: usbProdID = USB_II_16_DEV; break;
		case 6: usbProdID = USB_RO_16_DEV; break;
		case 7: usbProdID = USB_IIRO_8_DEV; break;
		case 8: usbProdID = USB_II_8_DEV; break;
		case 9: usbProdID = USB_RO_8_DEV; break;
		};
		
		CFDictionarySetValue(dictionary, CFSTR(kUSBProductID),
								CFNumberCreate(kCFAllocatorDefault, kCFNumberSInt32Type, &usbProdID));
		
		
		
		do
		{
			gDevFound = false;
			
			dictionary = (CFMutableDictionaryRef) CFRetain(dictionary);
			
			IOServiceAddMatchingNotification(gNotifyPort,
												kIOFirstMatchNotification,
												dictionary,
												devAdded,
												NULL,
												&gAddIter);
												
			devAdded(NULL, gAddIter);
			
			if (gDevFound) //add notification for removal
			{
				dictionary = (CFMutableDictionaryRef) CFRetain(dictionary);
				
				deviceList[gNumDevices].prodID = usbProdID;
				
				IOServiceAddMatchingNotification(gNotifyPort,
													kIOTerminatedNotification,
													dictionary,
													devRemoved,
													NULL,
													&deviceList[gNumDevices].removedIter);
													
				devRemoved(NULL, deviceList[gNumDevices].removedIter);
				gNumDevices ++;
				
				deviceList[gNumDevices - 1].connected = true;
				queryDeviceInfo(gNumDevices - 1, NULL, 0, NULL, &(deviceList[gNumDevices - 1].DIOBytes),
								&(deviceList[gNumDevices - 1].counters));
				
				
				if (gNumDevices == MAX_DEVICES - 1)
				{
					CFRelease(dictionary);
					return gNumDevices;
				}
			}
		}while (gDevFound == true);
								
		
	
	}
	
	//at the end of all this we have one extra reference to the dictionary so we need to release it
	CFRelease(dictionary);
	
	return gNumDevices;
}
